/* common.c - common parts of client/server that use the socket interface to tcp
 *		 This pair is modelled on rping -c.
 *		 They both accept a number of command line options.
 *		 The client then  repeatedly sends messages to the server,
 *		 and reads back the replies
 *		 Modified for exs API.
 *  on Solaris this program must be linked with -lnsl and -lsocket
 */

/*
 * Copyright 2007-2016 University of New Hampshire Interoperability Laboratory
 * All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING.GPL in the main directory of this source tree, or the
 * BSD license, available from the file COPYING.BSD in the main directory
 * of this source tree.
 *
 * This material is based upon work supported by the National Science
 * Foundation under Grants No. OCI-1127228 and DGE-0913620.
 *
 * Any opinions, findings, and conclusions or recommendations
 * expressed in this material are those of the authors and do not
 * necessarily reflect the views of the National Science Foundation.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <unistd.h>
#include <errno.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <ctype.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <time.h>
#include <netdb.h>

#include "utilities.h"
#include "latenexs_common.h"

struct timespec	start_time;
struct sockaddr_in6 server_addr;
struct sockaddr_in6 client_addr;


/*	-a: server_name == IP address of server	*/
char			*server_name;

/*	-b: nbufs == number of simultaneously active buffers	*/
int			nbufs = DEFAULT_NBUFS;

/*	-c: count == number of transactions	*/
int			count = DEFAULT_COUNT;

/*	-g: sg_elements == number of scatter_gather elements in exs_sendfile()*/
int			sg_elements = 0;

/*	-l: n_listener_loops == maximum number of connections to accept */
int			n_listener_loops = DEFAULT_N_LISTENER_LOOPS;

/*	-p: server_port == port number of server */
char			*server_port;

/*	-s: size == number of bytes per transaction	*/
int			size = DEFAULT_SIZE;

/*	-u: unregistered_buffers_flag = 1 to use unregistered buffers for I/O */
int			unregistered_buffers_flag = 0;

/*	-k: set EXS_FD_SP_COPY_UNREG_ON_SEND
 *				to copy immediate data into prereg buf
 *	-x: iset EXS_FD_BYPASS_PARAM_CHECKING to skip exs_xxx parameter checks
 *	-m: small_packet_max_size = maximum size of immediate data
 *	-F: completion thread CPU
 */
struct conn_params	our_conn_params;

/*	Currently not used
 */
int			recv_flags = 0;

/*	Currently not used
 */
int			send_flags = 0;

int			socktype = SOCK_SEQPACKET;

int use_registered_memory = 1;	/* default use registered mem */


void
print_latency( char *who, int count, int secs, int usecs )
{
	double	dtime, dmessages;

	fprintf(stderr, "%s finished %d messages in %d.%06d seconds\n",
		who, count, secs, usecs);

	dtime = secs;
	dtime *= MEGA_UNITS;
	dtime += usecs;
	dmessages = ((double)count);
	if (dmessages != 0) {
		fprintf(stderr, "%s: %.3f microseconds/message round-trip\n",
			who, dtime/dmessages);
	}
}	/* print_latency */


int
setup_buffers(const char *who, char **in_mess, exs_mhandle_t *in_mhandle,
				char **out_mess, exs_mhandle_t *out_mhandle)
{
	int ret = -1;

	*out_mess = calloc(size + 1, sizeof(char));
	if (!(*out_mess)) {
		fprintf(stderr, "%s calloc out_mess: %s\n",
			who, strerror(errno));
		goto out;
	}

	*in_mess = calloc(size + 1, sizeof(char));
	if (!(*in_mess)) {
		fprintf(stderr, "%s calloc in_mess: %s\n",
			who, strerror(errno));
		goto out1;
	}

	if (use_registered_memory) {
		/* need to register one block of memory that we send to server
		 * and another block that we get back from the server
		 */
		*out_mhandle = exs_mregister(*out_mess,
						size, EXS_MRF_RECV_DISABLE);
		if (*out_mhandle == EXS_MHANDLE_INVALID) {
			fprintf(stderr, "%s exs_mregister out message buffer: "
				"%s\n", who, strerror(errno));
			goto out2;
		}
		TRACE_DBG(LOG_APPLICATION,
			"%s %s: registered %d bytes of out memory at %p, handle %p",
			time_prefix(), who, size, out_mess, out_mhandle);

		*in_mhandle = exs_mregister(*in_mess, size, 0);
		if (*in_mhandle == EXS_MHANDLE_INVALID) {
			fprintf(stderr, "%s exs_mregister in message buffer: "
				"%s\n", who, strerror(errno));
			goto out3;
		}
		TRACE_DBG(LOG_APPLICATION,
			"%s %s: registered %d bytes of in memory at %p, handle %p",
			time_prefix(), who, size, in_mess, in_mhandle);
	} else {/* using unregistered memory */
		*out_mhandle = EXS_MHANDLE_UNREGISTERED;
		*in_mhandle = EXS_MHANDLE_UNREGISTERED;
	}
	ret = 0;
	goto out;

out3:
	exs_mderegister(*out_mhandle, 0);
	*out_mhandle = EXS_MHANDLE_INVALID;
out2:
	free(*in_mess);
	*in_mess = NULL;
out1:
	free(*out_mess);
	*out_mess = NULL;
out:
	return ret;
}	/* setup_buffers */


void
startup(int argc, char *argv[], int is_client)
{
	int		n;
	unsigned int	app_debug_level;
	unsigned int	util_debug_level;
	unsigned int	len;
	char	*ptr, c;
	char		*who;

	/* ignore sigpipe signals -- handle the error synchronously */
	no_sigpipe();

	/* initialize the exs API */
	if (exs_init(EXS_VERSION1) < 0) {
		perror("exs_init");
		exit(EXIT_FAILURE);
	}

	/* print exs counter statistics whenever we exit */
	atexit(exs_dump_counters);

	memset(&server_addr, 0, sizeof(server_addr));
	memset(&client_addr, 0, sizeof(client_addr));
	server_name = NULL;
	server_port = DEFAULT_PORT;
	count = DEFAULT_COUNT;
	size = DEFAULT_SIZE;
	app_debug_level = 0U;
	util_debug_level = 0U;
	our_conn_params.fd_flags = 0;
	our_conn_params.small_packet_max_size = INT_MAX;
	our_conn_params.max_inline = INT_MAX;
	our_conn_params.comp_cpu = INT_MAX;
	our_conn_params.max_credits = INT_MAX;
	our_conn_params.stream_buf_size = INT_MAX;

	if (is_client)
		who = "client";
	else
		who = "server";

	while ((c = getopt(argc, argv, OPTIONS)) != EOF) {
		switch (c) {
		case 'n':
		case 'N':
			/* turn off pseudo-nagle algorithm */
			our_conn_params.fd_flags |= EXS_FD_NODELAY;
			break;
		case 'x':
		case 'X':
			/* use "expert mode" to bypass parameter checking */
			n = exs_get_global_flags();
			exs_set_global_flags(n | EXS_GLOBAL_BYPASS_PARAM_CHECKING);
			break;
		case 'k':
		case 'K':
			/* copy unregistered immediate data into send buffer */
			/* ignored if -r is set */
			our_conn_params.fd_flags |= EXS_FD_SP_COPY_UNREG_ON_SEND;
			break;
		case 'm':
		case 'M':
			/* set max_size of "small packet" immediate data */
			errno = 0;
			n = strtol(optarg, &ptr, 0);
			if (ptr == optarg || errno != 0 || n < 0
			|| n >= (64 * 1024 * 1024)
			|| strspn(ptr, WHITE_SPACE) != strlen(ptr)) {
				fprintf(stderr, "\tillegal -%c value "
					"\"%s\"\n", c, optarg);
				exit(EXIT_FAILURE);
			}
			our_conn_params.small_packet_max_size = n;
			break;
		case 'u':
		case 'U':
			/* use unregistered memory, not registered memory */
			use_registered_memory = 0;
			break;
		case 'b':
		case 'B':
			/* number simultaneously active buffers */
			errno = 0;
			n = strtol(optarg, &ptr, 0);
			if (ptr == optarg || errno != 0 || n <= 0
			|| strspn(ptr, WHITE_SPACE) != strlen(ptr)) {
				fprintf(stderr, "\tillegal -%c value "
					"\"%s\"\n", c, optarg);
				exit(EXIT_FAILURE);
			}
			nbufs = n;
			break;
		case 'l':
		case 'L':
			/* number listener loops */
			errno = 0;
			n = strtol(optarg, &ptr, 0);
			if (ptr == optarg || errno != 0 || n == 0
			|| strspn(ptr, WHITE_SPACE) != strlen(ptr)) {
				fprintf(stderr, "\tillegal -%c value "
					"\"%s\"\n", c, optarg);
				exit(EXIT_FAILURE);
			}
			n_listener_loops = n;
			break;
		case 'D':
			n = exs_get_global_flags();
			exs_set_global_flags(n | EXS_GLOBAL_TRACE);
			util_debug_level++;
		case 'd':
			app_debug_level++;
			break;
		case 'a':
		case 'A':
			/* IP name or address of server */
			len = strlen(optarg);
			if (len <= 0 || strspn(optarg, WHITE_SPACE) == len)
				break;
			server_name = malloc(len+1);
			if (server_name == NULL) {
				perror("malloc for -a");
				break;
			}
			strcpy(server_name, optarg);
			break;
		case 'p':
		case 'P':
			/* port number */
			len = strlen(optarg);
			if (len <= 0 || strspn(optarg, WHITE_SPACE) == len)
				break;
			server_port = malloc(len+1);
			if (server_port == NULL) {
				perror("malloc for -p");
				break;
			}
			strcpy(server_port, optarg);
			break;
		case 'c':
		case 'C':
			/* count of number of iterations */
			errno = 0;
			n = strtol(optarg, &ptr, 0);
			if (ptr == optarg || errno != 0 || n <= 0
			|| strspn(ptr, WHITE_SPACE) != strlen(ptr)) {
				fprintf(stderr, "illegal -%c value \"%s\", "
					"ignored\n", c, optarg);
				break;
			}
			count = n;
			break;
		case 's':
		case 'S':
			/* size of data block sent in message */
			errno = 0;
			n = strtol(optarg, &ptr, 0);
			if (ptr == optarg || errno != 0 || n <= 0
			|| strspn(ptr, WHITE_SPACE) != strlen(ptr)) {
				fprintf(stderr, "illegal -%c value \"%s\", "
					"ignored\n", c, optarg);
				break;
			}
			size = n;
			break;
		case 'E':
			/* force busy-polling instead of event notification */
			our_conn_params.fd_flags |= EXS_FD_BUSYPOLL;
			break;
		case 'F':
			errno = 0;
			n = strtol(optarg, &ptr, 0);
			if (ptr == optarg || errno != 0 || n <= 0
			|| strspn(ptr, WHITE_SPACE) != strlen(ptr)) {
				fprintf(stderr, "illegal -%c value "
					"\"%s\"\n", c, optarg);
				exit(EXIT_FAILURE);
			}
			our_conn_params.comp_cpu = n;
			break;
		case 'r':
		case 'R':
			errno = 0;
			n = strtol(optarg, &ptr, 0);
			if (ptr == optarg || errno != 0 || n <= 0
			|| strspn(ptr, WHITE_SPACE) != strlen(ptr)) {
				fprintf(stderr, "illegal -%c value "
					"\"%s\"\n", c, optarg);
				exit(EXIT_FAILURE);
			}
			our_conn_params.max_credits = n;
			break;
		case 't':
		case 'T':
			socktype = SOCK_STREAM;
			break;
		case 'z':
		case 'Z':
			errno = 0;
			n = strtol(optarg, &ptr, 0);
			if (ptr == optarg || errno != 0 || n < 0
						|| strspn(ptr, WHITE_SPACE)
							!= strlen(ptr)) {
				fprintf(stderr, "illegal -%c value \"%s\"\n", c, optarg);
				exit(EXIT_FAILURE);
			}
			our_conn_params.stream_buf_size = n;
			break;
		case 'h':
		case 'H':
			printf("Usage: %s options ...\n", who);
			printf("-a <server's ip-address or dns-name>\n");
			printf("-b <number>\tNumber of simulataneously active buffers (default %d)\n",
								DEFAULT_NBUFS);
			printf("-c <number>\tNumber of iterations (default %d)\n",
								DEFAULT_COUNT);
			printf("-d\tEnable basic debug trace output\n");
			printf("-D\tEnable all debug trace output\n");
			printf("-E\tUse busy-polling instead of event notification\n");
			printf("-F <number>\tPin completion-processing threads to CPU core\n");
			printf("-k\tCopy memory to/from preregistered buffers\n");
			printf("-l\tNumber of connections to accept (default %d)\n",
						DEFAULT_N_LISTENER_LOOPS);
			printf("-m <number>\tMaximum size of preregistered buffers for -k\n");
			printf("-n\tTurn off pseudo-Nagle algorithm\n");
			printf("-p <number>\tServer's port number (default %s)\n",
								DEFAULT_PORT);
			printf("-r <number>\tMaximum number of credits to negotiate\n");
			printf("-s <number>\tSize of messages in bytes (default %d)\n",
								DEFAULT_SIZE);
			printf("-t\tUse SOCK_STREAM mode; not SOCK_SEQPACKET\n");
			printf("-u\tDo not register memory\n");
			printf("-x\tBypass parameter checking\n");
			printf("-z\tMaximum size of stream recv buffer\n");
			exit(EXIT_SUCCESS);
			break;
		default:
			fprintf(stderr, "unknown option '%c', ignored\n", c);
			break;
		}	/* switch */
	}	/* while */

	set_log_level(LOG_UTILITIES, util_debug_level);
	set_log_level(LOG_APPLICATION, app_debug_level);

	fprintf(stderr, "%s server_name: \"%s\"\n", time_prefix(), server_name);
	fprintf(stderr, "%s server_port: \"%s\"\n", time_prefix(), server_port);
	fprintf(stderr, "%s count: %d\n", time_prefix(), count);
	fprintf(stderr, "%s size: %d\n", time_prefix(), size);
}	/* startup */


/* create a time-stamp for debug trace printing in global time_string array */

static char	time_string[32];

char *
time_prefix(void)
{
	struct timespec	cur_tval;

	clock_gettime(CLOCK_EXS, &cur_tval);
	if(start_time.tv_nsec > cur_tval.tv_nsec) {
		cur_tval.tv_nsec += 1000000000;
		cur_tval.tv_sec -= 1;
	}
	cur_tval.tv_nsec -= start_time.tv_nsec;
	cur_tval.tv_sec -= start_time.tv_sec;
	snprintf(time_string, 32, "%3d.%09ld:", (int)cur_tval.tv_sec,
		cur_tval.tv_nsec);
	return time_string;
}	/* time_prefix */

/* vi: set autoindent tabstop=8 shiftwidth=8 : */
