/*	evt_queue.c	*/

/*
 * Copyright 2007-2016 University of New Hampshire Interoperability Laboratory
 * All rights reserved.
 *
 * This file is part of the UNH EXS implementation of the
 * Consortium Specification: Extended Sockets API (ES-API) Issue 1.0,
 * Document Number C050 published by The Open Group, January 2005,
 * and copyright 2005 by The Open Group.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING.GPL in the main directory of this source tree, or the
 * BSD license, available from the file COPYING.BSD in the main directory
 * of this source tree.
 *
 * This material is based upon work supported by the National Science
 * Foundation under Grants No. OCI-1127228 and DGE-0913620.
 *
 * Any opinions, findings, and conclusions or recommendations
 * expressed in this material are those of the authors and do not
 * necessarily reflect the views of the National Science Foundation.
 */

/*
 * The UNH IOL implementation of the Extended Sockets API (ES-API) Issue 1.0
 *
 * defined in:	The Interconnect Software Consortium Issue 1.0
 *		ISBN: 1-931624-52-6
 *		Published by The open Group, January 2005.
 *
 * all page references in this source code are to pages in that document.
 */

/*
 * Most of the coding in this file done by:
 *	Bob Russell
 *	rdr@iol.unh.edu
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <semaphore.h>
#include <signal.h>
#include <time.h>
#include <pthread.h>
#include <limits.h>

#include <atomic.h>
#include <evt_queue.h>
#include <exs_private.h>

/* default "depth" of event queue */
#define OUR_DEFAULT_QUEUE_DEPTH 64

/* number of slots per (re)allocation in queue_table */
#define	QUEUE_CHUNK_SIZE	64

/* table of qhandles that point to all created struct evt_queue */
static struct evt_queue **queue_table = NULL;

/* number of slots currently allocated to queue_table */
static unsigned long queue_table_limit = 0;

/* lowest index to free slot in queue_table (0 unused) */
static unsigned long lowest_available_qnumb = 1;

/* protect all access to queue_table */
static pthread_rwlock_t	qt_lock = PTHREAD_RWLOCK_INITIALIZER;


/* Returns the largest integer less than or equal to the logarithm base 10 of
 * the given integer.  This is equivalent to (unsigned)floor(log10(v)) if v is
 * nonzero.  For efficiency reasons, we return 0 if v is 0.  The returned value
 * is guaranteed to be small and nonnegative, hence it is returned as an
 * unsigned integer. */
static inline unsigned int
our_log10(unsigned long long v)
{
	unsigned int ret = 0;
	while ((v /= 10) != 0) {
		ret++;
	}
	return ret;
}	/* our_log10 */


/* called locally by _evt_queue_init(), _swap_evt_to_full_evt_queue(),
 *		and _post_event_to_full_evt_queue()
 * all these are local, so qptr is valid
 */
static exs_event_t *
alloc_exs_event(void)
{
	exs_event_t	*evt_ptr;

	evt_ptr = EXS_CALLOC(1, sizeof(struct exs_event));
	if (evt_ptr == NULL)
	{
		OUR_PERROR(errno, "alloc_exs_event");
	}
	return evt_ptr;
}	/* alloc_exs_event */


struct evt_queue *
_evt_queue_init_internal(unsigned int depth)
{
	int			ret;
	struct evt_queue 	*qptr;
	struct exs_event	*evt_ptr;

	/* allocate the evt_queue structure */
	qptr = EXS_CALLOC(1, sizeof(struct evt_queue));
	if (qptr == NULL)
		return NULL;

	TRACE_DBG("allocated qptr %p, %zu bytes",
		qptr, sizeof(struct evt_queue));

	/* initialize the full list variables */
	exs_p_list_init(&qptr->full_list, OUR_LIST_BLOCKING, depth);

	/* initialize the empty list variables */
	exs_p_list_init(&qptr->empty_list, 0, depth);

	if (depth == 0) {
		/* user wants use to set a default depth */
		depth = OUR_DEFAULT_QUEUE_DEPTH;
	}

	qptr->depth = qptr->nitems = depth;

	TRACE_DBG("set qptr %p depth %u",
		qptr, depth);

	/* allocate all the "struct exs_event" objects on this queue */
	while (depth > 0) {
		if ((evt_ptr = alloc_exs_event()) == NULL)
		{
			ret = errno;
			goto err3;
		}
		exs_p_list_add_tail(&qptr->empty_list, evt_ptr);
		depth--;
	}

	/* successful, return pointer to the new "struct evt_queue"*/
	qptr->qnumb = 0;
	qptr->wait = EXS_WAIT_ADAPTIVE;
	qptr->max_spins = 500;
	atomic_store(&qptr->refcount, 1);
	return qptr;

	/* else fall through into error recovery */
err3:
	while ((evt_ptr = exs_p_list_del_head(&qptr->empty_list))) {
		TRACE_DBG("qptr %p deallocating empty event %p",
			qptr, evt_ptr);
		EXS_FREE(evt_ptr);
	}

	EXS_FREE(qptr);
	errno = ret;
	return NULL;
}	/* _evt_queue_init_internal */


/* create and initialize a new event queue capable of handling "depth" events
 * Called from our_get_from_queue_cache(), exs_qcreate(), exs_blocking_close()
 *
 * Returns	pointer to the new queue on success
 *		NULL on failure, in which case global errno has error code
 */
struct evt_queue *
_evt_queue_init( unsigned int depth )
{
	int			ret;
	struct evt_queue 	*qptr, **newtable;
	unsigned long		new_qnumb, newlimit;

	TRACE_DBG("start, queue_table %p size %lu lowest %lu",
		queue_table, queue_table_limit,
		lowest_available_qnumb);

	/* allocate the evt_queue structure */
	qptr = _evt_queue_init_internal(depth);

	/* now add this newly-constructed qptr to queue table */
	OUR_RWLOCK_WRLOCK(qt_lock);

	for (new_qnumb = lowest_available_qnumb; new_qnumb < queue_table_limit;
						new_qnumb++) {
		if (queue_table[new_qnumb] == NULL) {
			/* found a free slot, use it */
			TRACE_DBG("found free slot %lu between lowest %lu and limit %lu",
				new_qnumb, lowest_available_qnumb,
				queue_table_limit);
			break;
		}
	}	/* for */
	ret = 0;
	if (new_qnumb >= queue_table_limit) {
		/* table is full, try to increase its size */
		TRACE_DBG("queue_table full, old base %p, old limit %lu",
			queue_table, queue_table_limit);

		/* if queue_table is NULL, realloc == alloc,
		 *	new_qnumb = lowest = 1, limit = 0;
		 * else new_qnumb = limit = k*CHUNK_SIZE
		 */
		newlimit = queue_table_limit + QUEUE_CHUNK_SIZE;
		newtable = realloc(queue_table,newlimit*sizeof(*newtable));
		if (newtable == NULL) {
			/* can't increase size, don't create new queue */
			qptr = _evt_queue_unref(qptr, 0, 0);
			ret = -1;
		} else {
			/* NULL out new space just appended to end of table */
			memset(&newtable[queue_table_limit], 0,
				QUEUE_CHUNK_SIZE * sizeof(*newtable));

			/* update static global table parameters */
			queue_table = newtable;
			queue_table_limit = newlimit;
			TRACE_DBG("(re)alloc ok, new queue_table %p new limit %lu",
				queue_table, queue_table_limit);
		}
	}
	if (ret == 0) {
		/* new_qnumb is lowest index in queue_table_limit of free slot*/
		qptr->qnumb = new_qnumb;
		queue_table[new_qnumb] = qptr;
		TRACE_DBG("queue_table[%lu] %p",
			new_qnumb, qptr);
		/* update lowest_available_qnumb
		 * (which may force it off current end of table)
		 */
		lowest_available_qnumb = new_qnumb + 1;
	}

	OUR_RWLOCK_UNLOCK(qt_lock);

	if (ret == 0) {
		/* successful, return pointer to the new "struct evt_queue"*/
		atomic_store(&qptr->refcount, 1);
		TRACE_DBG("leave, queue_table %p limit %lu lowest %lu qnumb %lu ret %p",
			queue_table,
			queue_table_limit, lowest_available_qnumb,
			new_qnumb, qptr);
		return qptr;
	} else {
		return NULL;
	}
}	/* _evt_queue_init */


static void
evt_queue_destroy(struct evt_queue *qptr)
{
	struct exs_event *evt_ptr;

	/* qptr is valid and has been removed from queue table
	 * free up its parts and then free it
	 */
	while ((evt_ptr = exs_p_list_del_head(&qptr->full_list))) {
		TRACE_DBG("qptr %p deallocating %zu-byte event %p type %d",
			qptr, sizeof(exs_event_t), evt_ptr,
			evt_ptr->exs_evt_type);
		EXS_FREE(evt_ptr);
	}
	exs_p_list_destroy(&qptr->full_list);

	while ((evt_ptr = exs_p_list_del_head(&qptr->empty_list))) {
		TRACE_DBG("qptr %p deallocating empty event %p",
			qptr, evt_ptr);
		EXS_FREE(evt_ptr);
	}
	exs_p_list_destroy(&qptr->empty_list);

	EXS_FREE(qptr);
}	/* evt_queue_destroy */


/* called from exs_qdelete()
 * returns 0 if all ok, -1 on error with global errno set
 *
 * This takes exs_qhandle_t so that it can do the translation, validity check,
 * NULLing of the pointer, and decrement of the refcount all under the
 * protection of qt_lock.  Otherwise, a thread could get in between and
 * re-increment the reference count and/or npending when this function thinks it
 * can go ahead and delete the queue.
 *
 * This function removes the queue from the table but only decrements its
 * refcount, so if the refcount is still positive, the last function out will
 * destroy the event queue.
 */
int
_evt_queue_deinit( exs_qhandle_t qhandle )
{
	struct evt_queue	*qptr;
	unsigned long		qnumb;
	int			ret;
	unsigned int		new_refcount = 1;

	/* new_refcount initialized to non-zero value so that if there is an
	 * error, it will still test non-zero and then we avoid doing the
	 * destroy */

	OUR_RWLOCK_WRLOCK(qt_lock);

	if ((qnumb = (unsigned long)qhandle) >= queue_table_limit
				|| (qptr = queue_table[qnumb]) == NULL
				|| qptr->qnumb != qnumb) {
		qptr = NULL;	/* qnumb is not valid */
		errno = EINVAL;	/* Invalid argument */
		ret = -1;
		goto errout;
	} else if (atomic_fetch(&qptr->npending) > 0) {
		/* still have pending async tasks */
		TRACE_ERR("qptr %p has %d tasks pending; returning EBUSY",
				(void *)qptr, atomic_fetch(&qptr->npending));
		ret = -1;
		errno = EBUSY;
		assert(atomic_fetch(&qptr->refcount) != 0);
		goto errout;
	}

	/* ok, NULL out the table and update lowest available */
	queue_table[qptr->qnumb] = NULL;
	if (lowest_available_qnumb > qptr->qnumb)
		lowest_available_qnumb = qptr->qnumb;
	exs_p_list_unblock(&qptr->full_list);
	TRACE_DBG("DECREMENT qptr %p refcount", (void *)qptr);
	/* Decrement for exs_qdelete() call *and* initial exs_qcreate() */
	new_refcount = atomic_uint_sub(&qptr->refcount, 1);
	ret = 0;

errout:
	OUR_RWLOCK_UNLOCK(qt_lock);

	if (!new_refcount) {
		/* We are the last ones out */
		TRACE_DBG("event queue qptr %p refcount reached 0",
								(void *)qptr);
		assert(atomic_fetch(&qptr->npending) == 0);
		evt_queue_destroy(qptr);
	} else {
		/* The last one to decrement new_refcount will do the final
		 * destroy */
		TRACE_DBG("qptr %p refcount still %u",
				(void *)qptr, new_refcount);
	}

	TRACE_DBG("qptr %p returning %d",
			qptr, ret);

	return ret;
}	/* _evt_queue_deinit */


/* Change the wait type of the queue.  Can only be done prior to using the queue
 * for the first time. */
int
_set_evt_queue_wait(struct evt_queue *qptr, int *new_wait)
{
	if (atomic_fetch(&qptr->npending) > 0 || exs_p_list_size(&qptr->full_list) != 0) {
		errno = EINVAL;
		return -1;
	}

	switch (*new_wait) {
	case EXS_WAIT_ADAPTIVE: /* fall through */
	case EXS_WAIT_NOTIFY: /* fall through */
	case EXS_WAIT_BUSY_POLL:
		break;
	default:
		errno = EINVAL;
		return -1;
	}

	qptr->wait = *new_wait;
	qptr = _evt_queue_unref(qptr, 1, 0);
	return 0;
}	/* _set_evt_queue_wait */


/* disables signal generation for this queue
 * returns 0 if all ok, -1 with global errno set on any error
 */
int
_evt_queue_disable_signal( struct evt_queue *qptr )
{
	qptr->signal_number = 0;
	qptr = _evt_queue_unref(qptr, 1, 0);
	return 0;
}	/* _evt_queue_disable_signal */


/* Called from our_enqueue_event() if list was empty prior to enqueueing the
 * event, and from _evt_queue_enable_signal if the list is nonempty.
 *
 * If all other threads have the signal blocked, but the calling thread has the
 * signal unblocked, the signal will be delivered to the calling thread before
 * this function returns, which may lead to the signal being lost.
 *
 * Since UNH EXS is multithreaded, the *only* "safe" way of using signals is to
 * block the signal in all threads and have one single thread use sigwaitinfo(),
 * etc. to wait for the incoming signal. Even still, the developers consider
 * the signaling mechanism to be a misfeature of the ES-API and we cannot
 * guarantee that it will work in every case.
 *
 * While this function is currently just a wrapper around kill(), we are using
 * this wrapper in case it is decided that more code is needed to handle the
 * signal correctly in the future.
 *
 * Raises any non-zero signo
 * returns 0 if all ok, else -1 with errno set
 */
static inline int
our_raise_signal(int signo)
{
	/* We must call kill() and not raise() because raise() will only send
	 * the signal to the calling thread, which may not be the thread that
	 * is waiting on the signal. */
	return kill(getpid(), signo);
}	/* our_raise_signal */


/* enables signal generation for this queue whenever a new event is
 * enqueued on a previously empty queue
 * returns 0 if all ok, -1 and errno set on any error
 */
int
_evt_queue_enable_signal( struct evt_queue *qptr, int signal_number )
{
	int			ret;

	if (signal_number <= 0) {
		errno = EINVAL;
		return -1;
	}

	qptr->signal_number = signal_number;
	ret = 0;

	TRACE_DBG("DECREMENT qptr %p refcount", (void *)qptr);
	if (atomic_uint_sub(&qptr->refcount, 1) == 0) {
		evt_queue_destroy(qptr);
	} else if (!exs_p_list_empty(&qptr->full_list)) {
		ret = our_raise_signal(signal_number);
	}
	return ret;
}	/* _evt_queue_enable_signal */


static void
hist_record(struct evt_queue *qptr, unsigned long long spins)
{
#if !defined(ALLOW_EXS_STATISTICS)
	(void)qptr;
	(void)spins;
#else
	int log_spins = our_log10(spins);
	assert(0 <= log_spins && log_spins <= 20);
	++qptr->spin_count_hist[log_spins];
	if (spins > qptr->max_spin_count[log_spins]) {
		qptr->max_spin_count[log_spins] = spins;
	}
	if (!qptr->min_spin_count[log_spins]
			|| spins < qptr->min_spin_count[log_spins]) {
		qptr->min_spin_count[log_spins] = spins;
	}
#endif
}	/* hist_record */

static int
our_wait_indefinitely(struct evt_queue *qptr, struct exs_event **evt_ptr)
{
	unsigned long long max_spins, spins;
	struct exs_event *evt;
	int ret;

	switch (qptr->wait) {
	case EXS_WAIT_ADAPTIVE:
		max_spins = qptr->max_spins;
		if (max_spins > 0 && atomic_fetch(&qptr->npending) > 0) {
			spins = 0;
			do {
				evt = exs_p_list_del_head(
						&qptr->full_list);
				++spins;
			} while (!evt && !errno && spins < max_spins);
			hist_record(qptr, spins);
			if (evt) {
				/* Increase the number of times we spin if we
				 * got a completion event by spinning */
				qptr->max_spins++;
			}
		} else {
			evt = NULL;
		}
		if (!evt && !errno) {
#if defined(ALLOW_EXS_STATISTICS)
			++qptr->n_waits;
#endif
			/* Decrease the number of times we spin since we didn't
			 * get the event in time */
			if (qptr->max_spins > 0) {
				qptr->max_spins--;
			}
			evt = exs_p_list_del_head_or_wait(
						&qptr->full_list, NULL);
		}
		ret = 0;
		break;
	case EXS_WAIT_NOTIFY:
		evt = exs_p_list_del_head_or_wait(&qptr->full_list,
								NULL);
		ret = 0;
		break;
	case EXS_WAIT_BUSY_POLL:
		spins = 0;
		do {
			evt = exs_p_list_del_head(&qptr->full_list);
			++spins;
		} while (!evt && !errno);
		hist_record(qptr, spins);
		ret = 0;
		break;
	default:
		TRACE_ERR("Invalid wait type %d for event queue %p",
				qptr->wait, qptr);
		evt = NULL;
		errno = EINVAL;
		ret = -1;
		break;
	};

	*evt_ptr = evt;
	return ret;
}	/* our_wait_indefinitely */


/* returns in vector up to count events posted to qptr
 * returns the number actually copied into vector if all ok
 * returns -1 with errno set if any error
 *
 * Called from exs_qdequeue(), exs_accept(), exs_connect(), exs_send(),
 *		exs_recv(), exs_blocking_close()
 * caller gives us timeout in struct timeval (seconds and microseconds);
 * our clock function and timedwait use struct timespec (seconds and nanosecs);
 */
int
_evt_queue_dequeue_vector( struct evt_queue *qptr, exs_event_t *vector,
			int count, const struct timeval *timeout )
{
	int			n, kount, old_errno;
	struct exs_event	*evt_ptr;
	struct timespec		abs_delay;

	if (count <= 0) {
		n = -1;
		errno = EINVAL;
		goto errout;
	}

	/* do this as soon as arrive, before locking, to get good abs time */
	if (timeout != NULL) {
		clock_gettime(CLOCK_REALTIME, &abs_delay);
		abs_delay.tv_sec += timeout->tv_sec;
		abs_delay.tv_nsec += 1000 * timeout->tv_usec;
		if (abs_delay.tv_nsec >= 1000000000) {
			abs_delay.tv_nsec -= 1000000000;
			abs_delay.tv_sec += 1;
		}
	}

	TRACE_DBG("qptr %p, vector %p, count %d",
		qptr, vector, count);

	/* qptr is valid, see if queue has any events on it */
	n = 0;
	old_errno = errno;
	errno = 0;
	if (timeout) {
		if (timeout->tv_sec == 0 && timeout->tv_usec == 0) {
			/* Do a strict poll; do not go through any conditional
			 * waits, even with a delay of 0. */
			evt_ptr = exs_p_list_del_head(&qptr->full_list);
		} else {
			evt_ptr = exs_p_list_del_head_or_wait(&qptr->full_list,
								&abs_delay);
		}
	} else {
		our_wait_indefinitely(qptr, &evt_ptr);
	}
	if (!evt_ptr) {
		if (errno == 0 || errno == ETIMEDOUT) {
			n = 0;
			errno = old_errno;
		} else {
			n = -1;
		}
		goto errout;
	} else {
		memcpy(&vector[0], evt_ptr, sizeof(exs_event_t));
		n++;
		exs_p_list_add_tail(&qptr->empty_list, evt_ptr);
		TRACE_DBG("qptr %p move %zu-byte event %p type %d from full to empty list",
			qptr, sizeof(exs_event_t),
			evt_ptr, evt_ptr->exs_evt_type);
	}

	/* ret == 0, so should be an event on the queue now,
	 * return up to "count" of them */
	kount = count;
	while (--kount > 0 && (evt_ptr = exs_p_list_del_head(
						&qptr->full_list))) {
		memcpy(&vector[n], evt_ptr, sizeof(exs_event_t));
		n++;
		exs_p_list_add_tail(&qptr->empty_list, evt_ptr);
		TRACE_DBG("qptr %p move %zu-byte event %p type %d from full to empty list",
			qptr, sizeof(exs_event_t),
			evt_ptr, evt_ptr->exs_evt_type);
		if (--kount <= 0)
			break;
	}	/* list_for_each_safe */

errout:
	TRACE_DBG("qptr %p dequeue to vector %p returning %d",
		qptr, vector, n);

	return n;
}	/* _evt_queue_dequeue_vector */


/* if qptr's empty list is not itself empty,
 * then removes first event struct on that list and places pointer in *evt_ptr
 * else tries to allocate a new one and place a pointer in *evt_ptr,
 * else sets *evt_ptr to NULL and returns -ENOBUFS
 *
 * Also sets *evt_ptr to NULL if qptr is NULL, but returns 0
 *
 * Called from our_start_accept(), our_start_recv(),
 * our_start_send(), our_start_shutdown()
 */
int
_get_empty_evt_queue_event(struct evt_queue *qptr, struct exs_event **evt_ptr)
{
	int ret;

	assert(evt_ptr != NULL);
	if (!qptr) {
		*evt_ptr = NULL;
		return 0;
	}

	/* qptr is valid, use it */
	TRACE_DBG("qptr %p qn %p empty_list %p empty %d",
		(void *)qptr, (void *)qptr->qnumb, (void *)&qptr->empty_list,
		exs_p_list_empty(&qptr->empty_list));
	if ((*evt_ptr = exs_p_list_del_head(&qptr->empty_list))) {
		TRACE_DBG("qptr %p remove empty event %p from empty_list",
			qptr, (void *)*evt_ptr);
		ret = 0;
	} else {
		/* no empty event structs, try to construct a new one; otherwise
		 * tell the user we ran out of event structures */
		if ((*evt_ptr = EXS_CALLOC(1, sizeof(**evt_ptr)))) {
			ret = 0;
		} else {
			ret = -ENOBUFS;
		}
	}

	return ret;
}	/* _get_empty_evt_queue_event */


/* Adds evt_ptr to the queue and signals if the queue was empty.  Returns 0 if
 * the event was successfully enqueued, or -errno if there was an error adding
 * it to the queue.  Releases the reference to the queue unconditionally.
 *
 * May generate spurious signals but is guaranteed that if no signal was
 * generated, the queue was non-empty. */
static int
our_enqueue_event(struct evt_queue *qptr, struct exs_event *evt_ptr)
{
	int ret;

	TRACE_DBG("qptr %p add full event %p with type %s to full_list, signo %d",
		(void *)qptr, (void *)evt_ptr,
		exs_evt_type_to_string(evt_ptr->exs_evt_type),
		qptr->signal_number);
	assert(atomic_fetch(&qptr->npending) > 0);
	atomic_uint_sub(&qptr->npending, 1);
	ret = exs_p_list_add_tail(&qptr->full_list, evt_ptr);
	if (ret < 0) {
		return ret;
	} else if (ret == OUR_LIST_WAS_EMPTY) {
		our_raise_signal(qptr->signal_number);
	}
	qptr = _evt_queue_unref(qptr, 1, 0);
	return 0;
}	/* our_enqueue_event */


/* adds evt_ptr to end of queue
 * returns 0 if all ok, else -1 with errno set
 *
 * Called from our_fill_in_accept_event(), our_post_user_event(),
 * our_handle_send_eof(), our_socket_free()
 */
int
_put_full_evt_queue_event( struct evt_queue *qptr, exs_event_t *evt_ptr )
{
	int			ret;

	if (qptr->qnumb > 0) {
		assert(queue_table[qptr->qnumb] == qptr);
	}

	assert(atomic_fetch(&qptr->refcount) > 1);
	ret = our_enqueue_event(qptr, evt_ptr);
	if (ret < 0) {
		errno = -ret;
		ret = -1;
	}
	return ret;
}	/* _put_full_evt_queue_event */


/* adds evt_ptr back to empty list
 * returns 0 if all ok, else -1 with errno set
 *
 * Called from our_exs_recv_done(), our_last_block_sent()
 */
int
_put_empty_evt_queue_event( struct evt_queue *qptr, exs_event_t *evt_ptr )
{
	int			ret;

	if (qptr->qnumb > 0) {
		assert(queue_table[qptr->qnumb] == qptr);
	}

	assert(atomic_fetch(&qptr->refcount) > 1);
	ret = exs_p_list_add_tail(&qptr->empty_list, evt_ptr);
	if (ret < 0) {
		errno = -ret;
		ret = -1;
	}
	return ret;
}	/* _put_empty_evt_queue_event */


/*
 * on entry, qptr is known to be not NULL.
 * on entry, conn_pending->qptr is known to be not NULL.
 * on exit, conn_pending->qptr is always set to be NULL.
 *
 * Called from our_connect_returning(), our_undo_exs_socket(),
 *		and our_close_accepting()
 *
 * if all ok, gets a new exs_event_t item from qptr's empty list,
 * fills it in with event info.
 *
 * returns 0 if all ok, else -1 with errno set (although result is never used!)
 */
int
_post_event_to_full_evt_queue(struct pending_connect *conn_pending)
{
	struct evt_queue	*qptr;
	struct exs_event	*evt_ptr;
	int			ret;

	qptr = conn_pending->qhandle;
	conn_pending->qhandle = NULL;
	if (qptr->qnumb > 0) {
		assert(queue_table[qptr->qnumb] == qptr);
	}

	if (!(evt_ptr = exs_p_list_del_head(&qptr->empty_list))) {
		if((evt_ptr = alloc_exs_event()) == NULL) {
			/* errno already set to ENOMEM
			* event is never sent -- lost!
			*/
			ret = -1;
			goto err1;
		}
	}

	/* post new connect event in first item on empty list */
	/* see page 18 for these settings */

	evt_ptr->exs_evt_type = conn_pending->evt_type;
	evt_ptr->exs_evt_errno = conn_pending->err_numb;
	evt_ptr->exs_evt_ahandle = conn_pending->ahandle;
	evt_ptr->exs_evt_socket = conn_pending->sockno;

	/* move the (now full) event from empty to full list */
	assert(atomic_fetch(&qptr->refcount) > 1);
	ret = our_enqueue_event(qptr, evt_ptr);
	if (ret < 0) {
		errno = -ret;
		ret = -1;
	}
	goto out;
err1:
	qptr = _evt_queue_unref(qptr, 1, EVT_QUEUE_REF_PENDING);
out:
	return ret;
}	/* _post_event_to_full_evt_queue */


/* Called only from exs_qstatus()
 * Returns the current wait method for the given event queue
 */
void
_get_evt_queue_wait(struct evt_queue *qptr, int *result)
{
	*result = qptr->wait;
	qptr = _evt_queue_unref(qptr, 1, 0);
}	/* _get_evt_queue_wait */


/* Called only from exs_qstatus()
 * Returns a snapshot of the current statistics of the queue */
void
_get_evt_queue_statistics(struct evt_queue *qptr,
					struct exs_qhandle_statistics *stats)
{
#if !defined(ALLOW_EXS_STATISTICS)
	(void)qptr;
	(void)stats;
#else
	unsigned long *rbegin, *iter;

	stats->sleep_count = qptr->n_waits;
	stats->min_spin_count = qptr->min_spin_count[0];
	stats->target_spin_count = qptr->max_spins;
	iter = &qptr->max_spin_count[19];
	rbegin = &qptr->max_spin_count[0] - 1;
	do {
		stats->max_spin_count = *iter;
		iter--;
	} while (iter != rbegin && stats->max_spin_count == 0);

	memcpy(stats->spin_count_hist, qptr->spin_count_hist,
					sizeof(stats->spin_count_hist));
#endif
}	/* _get_evt_queue_statistics */


/* Called only from exs_qstatus()
 * Returns 0 if all ok and sets result, else -1 with errno set
 */
int
_get_evt_queue_depth( struct evt_queue *qptr, int *result )
{
	*result = qptr->depth;
	qptr = _evt_queue_unref(qptr, 1, 0);
	return 0;
}	/* _get_evt_queue_depth */


/* Called only from exs_qstatus()
 * Returns 0 if all ok and sets sig, else -1 with errno set
 */
int
_get_evt_queue_signal( struct evt_queue *qptr, struct exs_signal *sig)
{
	if (qptr->signal_number > 0) {
		sig->exs_signo = qptr->signal_number;
		sig->exs_sigstate = EXS_SIG_ENABLE;
	} else {
		sig->exs_signo = 0;
		sig->exs_sigstate = EXS_SIG_DISABLE;
	}
	qptr = _evt_queue_unref(qptr, 1, 0);
	return 0;
}	/* _get_evt_queue_signal */


/* if all ok, returns 0
 *	and sets in result number of events currently on qptr's full_list
 * else returns -1 with errno set
 *
 * Called only from exs_qstatus()
 */
int
_get_evt_queue_n_events( struct evt_queue *qptr, int *result )
{
	*result = exs_p_list_size(&qptr->full_list);
	qptr = _evt_queue_unref(qptr, 1, 0);
	return 0;
}	/* _get_evt_queue_n_events */


/* Decrements the reference count of the queue and returns NULL.  The caller
 * must no longer use its reference to this queue after calling this
 * function. */
struct evt_queue *
_evt_queue_unref(struct evt_queue *qptr, unsigned int n, int flags)
{
	assert(!(flags & ~EVT_QUEUE_REF_PENDING));
	assert(n > 0);
	assert(n <= atomic_fetch(&qptr->refcount));
	TRACE_DBG("DECREMENT qptr %p refcount", (void *)qptr);
	if (flags & EVT_QUEUE_REF_PENDING) {
		assert(n <= atomic_fetch(&qptr->npending));
		atomic_uint_sub(&qptr->npending, n);
	}
	if (atomic_uint_sub(&qptr->refcount, n) == 0) {
		evt_queue_destroy(qptr);
	}
	return NULL;
}	/* _evt_queue_unref */


/* basic conversion from user's qnumb to our qptr format
 * Called near the beginning of most exs_xxx() functions
 * to verify that the user's qptr parameter is valid.
 *
 * Returns true pointer to struct evt_queue if ok, else NULL and errno is set
 *
 * The reference count will be incremented by n
 * If flags & EVT_QUEUE_PENDING, npending will also be incremented by n
 *
 * Designed to return NULL (error) in 2 special cases:
 *	qh == NULL
 *	qh == EXS_QHANDLE_INVALID
 */
struct evt_queue *
_cnv_qnumb_to_ptr(exs_qhandle_t qn, unsigned int n, int flags)
{
	struct evt_queue *qptr;
	unsigned long qnumb;

	assert(!(flags & ~EVT_QUEUE_REF_PENDING));

	OUR_RWLOCK_RDLOCK(qt_lock);

	/* check that user's qn is valid */
	if ((qnumb = (unsigned long)qn) >= queue_table_limit
				|| (qptr = queue_table[qnumb]) == NULL
				|| qptr->qnumb != qnumb) {
		qptr = NULL;	/* qnumb is not valid */
		errno = EINVAL;	/* Invalid argument */
	} else {
		TRACE_DBG("INCREMENT qptr %p refcount", (void *)qptr);
		atomic_uint_add(&qptr->refcount, n);
		if (flags & EVT_QUEUE_REF_PENDING) {
			atomic_uint_add(&qptr->npending, n);
		}
	}

	OUR_RWLOCK_UNLOCK(qt_lock);

	TRACE_DBG("exs_qhandle %p return qptr %p",
		qn, qptr);
	return qptr;
}	/* _cnv_qnumb_to_ptr */

/* basic conversion from our qptr format to user's exs_qhandle_t
 * Called only from exs_qcreate() and exs_blocking_close()
 *
 * Returns user's pseudo-pointer if ok, else NULL (and errno is NOT set)
 */
exs_qhandle_t
_cnv_ptr_to_qnumb( struct evt_queue *qptr )
{
	unsigned long	qnumb;

	OUR_RWLOCK_RDLOCK(qt_lock);

	/* check that qptr is still valid */
	if ((qnumb = qptr->qnumb) <= 0
				|| qnumb >= queue_table_limit
				|| queue_table[qnumb] != qptr) {
		qnumb = 0;	/* qptr is not valid */
	}

	OUR_RWLOCK_UNLOCK(qt_lock);

	TRACE_DBG("qptr %p return exs_qhandle %p",
		qptr, (exs_qhandle_t)qnumb);
	return (exs_qhandle_t)qnumb;
}	/* _cnv_ptr_to_qnumb */

/* vi: set autoindent tabstop=8 shiftwidth=8 : */
