/*
 * Copyright 2007-2016 University of New Hampshire Interoperability Laboratory
 * All rights reserved.
 *
 * This file is part of the UNH EXS implementation of the
 * Consortium Specification: Extended Sockets API (ES-API) Issue 1.0,
 * Document Number C050 published by The Open Group, January 2005,
 * and copyright 2005 by The Open Group.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING.GPL in the main directory of this source tree, or the
 * BSD license, available from the file COPYING.BSD in the main directory
 * of this source tree.
 *
 * This material is based upon work supported by the National Science
 * Foundation under Grants No. OCI-1127228 and DGE-0913620.
 *
 * Any opinions, findings, and conclusions or recommendations
 * expressed in this material are those of the authors and do not
 * necessarily reflect the views of the National Science Foundation.
 */

#include "result.h"
#include "setup/exs_send_setup.h"
#include "setup/exs_client_setup.h"
#include "setup/exs_server_setup.h"
#include "setup/exs_qcreate_setup.h"

void *test_send(void *test_void_ptr)
{
	setup_fault_catcher(p, getpid());

	param *p = (param *) test_void_ptr;
	int test = p->test_number;

	/* EPERM: exs_init has not been called */
	if (test == 1) {
		p->test_value = "EPERM";
		p->test_name = all_send_tests[test - 1];
		int ret =
		    exs_send(p->fd, p->send_buf, p->SEND_BUFSIZE, p->flags,
			     p->qh, p->ah, p->mh);
		print_macro(test, p->test_value, p->test_name, ret, -1,
			    EPERM);
	}

	/* ENOTCONN: socket never connected */
	else if (test == 2) {
		p->test_value = "ENOTCONN";
		p->test_name = all_send_tests[test - 1];

		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);

		if (p->mode > 0)
			if (ok_exs_qcreate(p) < 0)
				exit(EXIT_FAILURE);

		if (ok_exs_socket(p) < 0)
			exit(EXIT_FAILURE);

		int ret =
		    exs_send(p->fd, p->send_buf, p->SEND_BUFSIZE, p->flags,
			     p->qh, p->ah, p->mh);

		print_macro(test, p->test_value, p->test_name, ret, -1,
			    ENOTCONN);
	}

	/* EBADF: bad fd */
	else if (test == 3) {
		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);
		if (ok_exs_socket(p) < 0)
			exit(EXIT_FAILURE);

		if (p->mode > 0)
			if (ok_exs_qcreate(p) < 0)
				exit(EXIT_FAILURE);
		p->test_value = "EBADF";
		p->test_name = all_send_tests[test - 1];
		int ret =
		    exs_send(-1, p->send_buf, p->SEND_BUFSIZE, p->flags,
			     p->qh, p->ah, p->mh);
		print_macro(test, p->test_value, p->test_name, ret, -1,
			    EBADF);
	}

	/* EINVAL: buffer is null */
	else if (test == 4) {
		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);
		if (ok_exs_socket(p) < 0)
			exit(EXIT_FAILURE);
		p->test_value = "EINVAL";
		p->test_name = all_send_tests[test - 1];
		int ret =
		    exs_send(p->fd, NULL, p->SEND_BUFSIZE, p->flags, p->qh,
			     p->ah, p->mh);
		print_macro(test, p->test_value, p->test_name, ret, -1,
			    EINVAL);
	}

	/* EINVAL: EXS_UNSIGNALED and EXS_BLOCK are both specified */
	else if (test == 5) {
		if (p->mode > 0)
			return 0;
		pid_t pid = fork();
		if (pid == 0) {
			exs_server(p);
			exit(EXIT_SUCCESS);
		} else {
			int fd = open(exs_test_fifo_1, O_RDONLY);

			if (ok_exs_init(p) < 0)
				exit(EXIT_FAILURE);
			if (ok_exs_socket(p) < 0)
				exit(EXIT_FAILURE);
			if (ok_exs_connect(p) < 0)
				exit(EXIT_FAILURE);

			close(fd);
			p->test_value = "EINVAL";
			p->test_name = all_send_tests[test - 1];
			int ret =
			    exs_send(p->fd, p->send_buf, p->SEND_BUFSIZE,
				     EXS_UNSIGNALED & EXS_BLOCK, p->qh,
				     p->ah, p->mh);
			print_macro(test, "EINVAL", p->test_name, ret, -1,
				    EINVAL);
			if (ok_exs_close(p) < 0)
				exit(EXIT_FAILURE);

			fd = open(exs_test_fifo_2, O_WRONLY);

			close(fd);
		}
	}

	/* EOPNOTSUPP: flags not supported */
	else if (test == 6) {
		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);
		if (ok_exs_socket(p) < 0)
			exit(EXIT_FAILURE);
		p->test_value = "EOPNOTSUPP";
		p->test_name = all_send_tests[test - 1];
		int ret =
		    exs_send(p->fd, p->send_buf, p->SEND_BUFSIZE, -1,
			     p->qh, p->ah, p->mh);
		print_macro(test, p->test_value, p->test_name, ret, -1,
			    EOPNOTSUPP);
	}

	/* EINVAL: qhandle invalid */
	else if (test == 7) {
		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);
		if (ok_exs_socket(p) < 0)
			exit(EXIT_FAILURE);
		p->test_value = "EINVAL";
		p->test_name = all_send_tests[test - 1];
		int ret =
		    exs_send(p->fd, p->send_buf, p->SEND_BUFSIZE, p->flags,
			     (exs_qhandle_t) - 1, p->ah, p->mh);
		print_macro(test, p->test_value, p->test_name, ret, -1,
			    EINVAL);
	}

	/* EINVAL: qhandle is NULL */
	else if (test == 8) {
		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);
		if (ok_exs_socket(p) < 0)
			exit(EXIT_FAILURE);
		p->test_value = "EINVAL";
		p->test_name = all_send_tests[test - 1];
		int ret =
		    exs_send(p->fd, p->send_buf, p->SEND_BUFSIZE, 0, NULL,
			     p->ah, p->mh);
		print_macro(test, p->test_value, p->test_name, ret, -1,
			    EINVAL);
	}

	/* EBUSY: no credits available */
	else if (test == 9) {

	}

	/* EPIPE: gracefull shut down for writing */
	else if (test == 10) {
		pid_t pid = fork();
		if (pid == 0) {
			int fd;
			if (p->mode == 0)
				fd = open(exs_test_fifo_1, O_RDONLY);

			if (ok_exs_init(p) < 0)
				exit(EXIT_FAILURE);
			if (p->mode > 0)
				if (ok_exs_qcreate(p) < 0)
					exit(EXIT_FAILURE);
			if (ok_exs_socket(p) < 0)
				exit(EXIT_FAILURE);
			if (ok_exs_connect(p) < 0)
				exit(EXIT_FAILURE);

			if (p->mode > 0)
				if (ok_exs_qdequeue(p) < 0)
					exit(EXIT_FAILURE);
			if (p->mode == 0)
				close(fd);

			int shut_ret =
			    exs_shutdown(p->fd, SHUT_WR, p->flags, p->qh,
					 p->ah);

			/** start exs_send **/
			p->test_value = "EPIPE";
			p->test_name = all_send_tests[test - 1];
			int ret =
			    exs_send(p->fd, p->send_buf, p->SEND_BUFSIZE,
				     p->flags, p->qh, p->ah, p->mh);
			print_macro(test, p->test_value, p->test_name, ret,
				    -1, EPIPE);

			/** end send **/

			if (p->mode == 0) {
				fd = open(exs_test_fifo_2, O_WRONLY);
				close(fd);
			}
			exit(EXIT_SUCCESS);
		} else {
			if (p->mode == 0)
				exs_server(p);
			if (p->mode > 0) {
				if (ok_exs_init(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_qcreate(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_socket(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_bind(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_listen(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_accept(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_mregister(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_qdequeue(p) < 0)
					exit(EXIT_FAILURE);
			}

		}
	} else if (test == 11) {	/*ECONNRESET: abruptly terminated */

	} else if (test == 12) {	/*ECANCELED: operation has been cancelled */

	} else if (test == 13) {	/*ECONNRESET: forcibly closed by a peer //TODO */

	} else if (test == 14) {	/*EFAULT: detected address was not vaild for buffer argument */

	} else if (test == 15) {	/*ENETDOWN: local interface is down */

	}

	/* OK_EXS_SEND */
	else if (test == 16) {
		p->test_name =
		    "exs_send(p->fd, buf, BUFSIZE, EXS_BLOCK, NULL, NULL, NULL)";
		pid_t pid = fork();
		if (pid == 0) {
			/** blocking client **/
			if (p->mode == 0) {
				int fd = open(exs_test_fifo_1, O_RDONLY);

				if (ok_exs_init(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_socket(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_connect(p) < 0)
					exit(EXIT_FAILURE);

				close(fd);

				p->test_value = "OK_EXS_SEND";
				p->test_name = all_send_tests[test - 1];
				int ret =
				    exs_send(p->fd, p->send_buf,
					     p->SEND_BUFSIZE, p->flags,
					     p->qh, p->ah, p->mh);

				if (ret < p->SEND_BUFSIZE) {
					print_macro(test, p->test_value,
						    p->test_name, ret, -1,
						    ENOTSUP);
				} else {
					print_macro(test, p->test_value,
						    p->test_name, ret,
						    p->SEND_BUFSIZE, 0);
				}

				if (ok_exs_close(p) < 0)
					exit(EXIT_FAILURE);

				fd = open(exs_test_fifo_2, O_WRONLY);

				close(fd);
			}
			/** non-blocking client **/
			if (p->mode > 0) {
				if (ok_exs_init(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_qcreate(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_socket(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_connect(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_qdequeue(p) < 0)
					exit(EXIT_FAILURE);

				/** send **/
				p->test_value = "OK_EXS_SEND";
				p->test_name = all_send_tests[test - 1];
				int ret =
				    exs_send(p->fd, p->send_buf,
					     p->SEND_BUFSIZE, p->flags,
					     p->qh, p->ah, p->mh);
				if (ok_exs_qdequeue(p) < 0)
					exit(EXIT_FAILURE);

				if (ret < 0) {
					print_macro(test, p->test_value,
						    p->test_name, ret, -1,
						    ENOTSUP);
				} else {
					print_macro(test, p->test_value,
						    p->test_name, ret, 0,
						    0);
				}
			}
			exit(EXIT_SUCCESS);
		} else {
			/** blocking server **/
			if (p->mode == 0) {

				if (ok_exs_init(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_socket(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_bind(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_listen(p) < 0)
					exit(EXIT_FAILURE);

				int fd = open(exs_test_fifo_1, O_WRONLY);

				if (ok_exs_accept(p) < 0)
					exit(EXIT_FAILURE);

				close(fd);

				if (ok_exs_close(p) < 0)
					exit(EXIT_FAILURE);

				fd = open(exs_test_fifo_2, O_RDONLY);

				close(fd);
			}
			/** non-blocking server **/
			if (p->mode > 0) {
				if (ok_exs_init(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_qcreate(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_socket(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_bind(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_listen(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_accept(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_qdequeue(p) < 0)
					exit(EXIT_FAILURE);
			}
		}
	}

    /** test number not valid **/
	else if (test != 0) {
		fprintf(stdout, "Invalid Test Number: %d\n", test);
	}
}

int main(int argc, char *argv[])
{
	execute(argc, argv, test_send, SEND_TESTS);
}
