/*
 * Copyright 2016 University of New Hampshire Interoperability Laboratory
 * All rights reserved.
 *
 * This file is part of the UNH EXS implementation of the
 * Consortium Specification: Extended Sockets API (ES-API) Issue 1.0,
 * Document Number C050 published by The Open Group, January 2005,
 * and copyright 2005 by The Open Group.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING.GPL in the main directory of this source tree, or the
 * BSD license, available from the file COPYING.BSD in the main directory
 * of this source tree.
 *
 * This material is based upon work supported by the National Science
 * Foundation under Grants No. OCI-1127228 and DGE-0913620.
 *
 * Any opinions, findings, and conclusions or recommendations
 * expressed in this material are those of the authors and do not
 * necessarily reflect the views of the National Science Foundation.
 */

#include "result.h"

static void test_setsockopt_ok(param *p, int level, int optname,
			       void *optval, socklen_t optlen)
{
	int ret;

	p->test_value = "OK_EXS_SETSOCKOPT";
	p->test_name = all_setsockopt_tests[p->test_number - 1];
	if (ok_exs_init(p) < 0)
		exit(EXIT_FAILURE);
	if (ok_exs_socket(p) < 0)
		exit(EXIT_FAILURE);
	ret = exs_setsockopt(p->fd, level, optname, optval, optlen);
	print_macro(p->test_number, p->test_value,
		    p->test_name, ret, 0, 0);
}

static void test_setsockopt_bad_length(param *p, int level, int optname,
					void *optval, socklen_t optlen)
{
	int ret;

	p->test_value = "EINVAL";
	p->test_name = all_setsockopt_tests[p->test_number - 1];
	if (ok_exs_init(p) < 0)
		exit(EXIT_FAILURE);
	if (ok_exs_socket(p) < 0)
		exit(EXIT_FAILURE);
	ret = exs_setsockopt(p->fd, level, optname, optval, optlen);
	print_macro(p->test_number, p->test_value, p->test_name,
		    ret, -1, EINVAL);
}

static void test_setsockopt_enoprotoopt(param *p, int level)
{
	int v, ret;

	p->test_value = "ENOPROTOOPT";
	p->test_name = all_setsockopt_tests[p->test_number - 1];
	if (ok_exs_init(p) < 0)
		exit(EXIT_FAILURE);
	if (ok_exs_socket(p) < 0)
		exit(EXIT_FAILURE);
	v = 1;
	ret = exs_setsockopt(p->fd, level, -1, &v, sizeof(v));
	print_macro(p->test_number, p->test_value, p->test_name,
		    ret, -1, ENOPROTOOPT);
}

static void *test_exs_setsockopt(void *test_void_ptr)
{
	param *p = (param *)test_void_ptr;
	uint32_t v32;
	int test = p->test_number;
	int v, ret;

	setup_fault_catcher(p, getpid());

	switch (test) {
	case 1:
		p->test_value = "EPERM";
		p->test_name = all_setsockopt_tests[test - 1];
		v = 1;
		ret = exs_setsockopt(p->fd, SOL_SOCKET, SO_REUSEADDR,
				     &v, sizeof(v));
		print_macro(p->test_number, p->test_value, p->test_name,
			    ret, -1, EPERM);
		break;
	case 2:
		p->test_value = "EBADF";
		p->test_name = all_setsockopt_tests[test - 1];
		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);
		v = 1;
		ret = exs_setsockopt(-1, SOL_SOCKET, SO_REUSEADDR,
				     &v, sizeof(v));
		print_macro(p->test_number, p->test_value, p->test_name,
			    ret, -1, EBADF);
		break;
	case 3:
		p->test_value = "EOPNOTSUPP";
		p->test_name = all_setsockopt_tests[test - 1];
		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);
		if (ok_exs_socket(p) < 0)
			exit(EXIT_FAILURE);
		v = 1;
		ret = exs_setsockopt(p->fd, -1, SO_REUSEADDR,
				     &v, sizeof(v));
		print_macro(p->test_number, p->test_value, p->test_name,
			    ret, -1, EOPNOTSUPP);
		break;
	case 4:
		test_setsockopt_enoprotoopt(p, SOL_SOCKET);
		break;
	case 5:
		test_setsockopt_enoprotoopt(p, SOL_EXS);
		break;
	case 6:
		v = 1;
		test_setsockopt_ok(p, SOL_SOCKET, SO_REUSEADDR, &v, sizeof(v));
		break;
	case 7:
		v32 = 64;
		test_setsockopt_ok(p, SOL_EXS, EXS_INLINE, &v32, sizeof(v32));
		break;
	case 8:
		v = 1;
		test_setsockopt_bad_length(p, SOL_SOCKET, SO_REUSEADDR, &v,
					   sizeof(v) - 1);
		break;
	case 9:
		v = 1;
		test_setsockopt_ok(p, SOL_SOCKET, SO_REUSEADDR,
				   &v, sizeof(v) + 1);
		break;
	case 10:
		v32 = 64;
		test_setsockopt_bad_length(p, SOL_EXS, EXS_INLINE, &v32,
					   sizeof(v32) - 1);
		break;
	case 11:
		v32 = 64;
		test_setsockopt_ok(p, SOL_EXS, EXS_INLINE,
				   &v, sizeof(v32) + 1);
		break;
	default:
		printf("Invalid Test Number: %d\n", test);
		break;
	}
} /* test_exs_setsockopt */

int main(int argc, char *argv[])
{
	execute(argc, argv, test_exs_setsockopt, SETSOCKOPT_TESTS);
} /* main */
